import os
import re
import shutil
import zipfile
from enum import Enum
from pathlib import Path

import requests
from bs4 import BeautifulSoup

from openpecha.core.layer import LayerEnum
from openpecha.serializers.serialize import Serialize
from openpecha.utils import load_yaml


class TsadraTemplateCSSClasses(Enum):
    cover_title = "credits-page_front-title"
    book_title = "tibetan-book-title"
    sub_title = "tibetan-book-sub-title"
    book_number = "tibetan-book-number"
    chapter = "tibetan-chapters"
    tsawa_inline = "tibetan-root-text"
    tsawa_verse = "tibetan-root-text-in-verse"
    citation_inline = "tibetan-citations"
    citation_verse = "tibetan-citations-in-verse"
    citation_prose = "tibetan-citations-prose"
    sabche = "tibetan-sabche1"
    yigchung = "tibetan-commentary-small"


class Tsadra_template:
    """Variables are important components of tsadra epub template."""

    # SP = Start_Payload
    span_EP = "</span>"
    para_EP = "</p>"
    ft = '<span class="front-title">'
    cover_title_SP = f'<span class="{TsadraTemplateCSSClasses.cover_title.value}">'
    book_title_SP = f'<span class="{TsadraTemplateCSSClasses.book_title.value}">'
    sub_title_SP = f'<span class="{TsadraTemplateCSSClasses.sub_title.value}">'
    book_number_SP = f'<p class="{TsadraTemplateCSSClasses.book_number.value}">{ft}'
    credit_page_SP = (
        '<p class="credits-page_epub-edition-line"><span class="credits-regular">'
    )
    author_SP = '<p class="text-author"><span class="front-page---text-titles">'
    chapter_SP = f'<span class="{TsadraTemplateCSSClasses.chapter.value}">'
    tsawa_inline_SP = f'<span class="{TsadraTemplateCSSClasses.tsawa_inline.value}">'
    tsawa_verse_SP = f'<span class="{TsadraTemplateCSSClasses.tsawa_verse.value}">'
    citation_inline_SP = (
        f'<span class="{TsadraTemplateCSSClasses.citation_inline.value}">'
    )
    citation_verse_SP = (
        f'<span class="{TsadraTemplateCSSClasses.citation_verse.value}">'
    )
    citation_prose_sp = (
        f'<span class="{TsadraTemplateCSSClasses.citation_prose.value}">'
    )
    sabche_SP = f'<span class="{TsadraTemplateCSSClasses.sabche.value}">'
    yigchung_SP = f'<span class="{TsadraTemplateCSSClasses.yigchung.value}">'
    footnote_marker_SP = '<span class="tibetan-footnote-marker"'
    footnote_EP = "</span></a>"
    footnote_reference_SP = '<span class="tibetan-footnote-reference"'

    toc_xpaths = {
        "book-number": f"//*[@class='{TsadraTemplateCSSClasses.book_number.value}']",
        "chapter": f"//*[@class='{TsadraTemplateCSSClasses.chapter.value}']",
        "sabche": f"//*[@class='{TsadraTemplateCSSClasses.sabche.value}' or @class='tibetan-sabche']",
    }
    toc_levels = {"1": "book-number", "2": "chapter", "3": "sabche"}
    book_title_Xpath = f"//*[@class='{TsadraTemplateCSSClasses.book_title.value}']"


class EpubSerializer(Serialize):
    """Epub serializer class for OpenPecha."""

    def __get_adapted_span(self, span, base_id):
        """Adapts the annotation span to base-text of the text

        Adapts the annotation span, which is based on volume base-text
        to text base-text.

        Args:
            span (dict): span of a annotation, eg: {start:, end:}
            base_id (str): id of vol, where part of the text exists.

        Returns:
            adapted_start (int): adapted start based on text base-text
            adapted_end (int): adapted end based on text base-text

        """
        adapted_start = span["start"] - self.text_spans[base_id]["start"]
        adapted_end = span["end"] - self.text_spans[base_id]["start"]
        return adapted_start, adapted_end

    def get_citation_sp(self, css_class_name):
        """Return start payload for citation according to class name

        Args:
            css_class_name (str): css class name to differentiate different types of citations

        Returns:
            str: citation start payload
        """
        start_payload = Tsadra_template.citation_inline_SP
        if css_class_name:
            css_class_enum = TsadraTemplateCSSClasses(css_class_name)
            if css_class_enum == TsadraTemplateCSSClasses.citation_verse:
                start_payload = Tsadra_template.citation_verse_SP
            elif css_class_enum == TsadraTemplateCSSClasses.citation_prose:
                start_payload = Tsadra_template.citation_prose_sp
        return start_payload

    def get_tsawa_sp(self, css_class_name):
        """Return start payload for tsawa according to class name

        Args:
            css_class_name (str): css class name to differentiate different types of tsawa

        Returns:
            str: tsawa start payload
        """
        start_payload = Tsadra_template.tsawa_inline_SP
        if css_class_name:
            css_class_enum = TsadraTemplateCSSClasses(css_class_name)
            if css_class_enum == TsadraTemplateCSSClasses.tsawa_verse:
                start_payload = Tsadra_template.tsawa_verse_SP
        return start_payload

    def apply_annotation(self, base_id, ann, uuid2localid):
        """Applies annotation to specific volume base-text, where part of the text exists.

        Args:
            base_id (str): id of vol, where part of the text exists.
            ann (dict): annotation of any type.

        Returns:
            None

        """
        only_start_ann = False
        start_payload = "("
        end_payload = ")"
        ann_type = LayerEnum(ann["type"])
        if ann_type == LayerEnum.correction:
            start_payload = "("
            end_payload = f',{ann["correction"]})'
        elif ann_type == LayerEnum.peydurma:
            start_payload = "#"
            only_start_ann = True
        elif ann_type == LayerEnum.error_candidate:
            start_payload = "["
            end_payload = "]"
        elif ann_type == LayerEnum.book_title:
            start_payload = Tsadra_template.book_title_SP
            end_payload = Tsadra_template.span_EP
        elif ann_type == LayerEnum.sub_title:
            start_payload = Tsadra_template.sub_title_SP
            end_payload = Tsadra_template.span_EP
        elif ann_type == LayerEnum.book_number:
            start_payload = Tsadra_template.book_number_SP
            end_payload = f"{Tsadra_template.span_EP}{Tsadra_template.para_EP}"
        elif ann_type == LayerEnum.author:
            start_payload = Tsadra_template.author_SP
            end_payload = f"{Tsadra_template.span_EP}{Tsadra_template.para_EP}"
        elif ann_type == LayerEnum.chapter:
            start_payload = Tsadra_template.chapter_SP
            end_payload = Tsadra_template.span_EP
        elif ann_type == LayerEnum.tsawa:
            css_class_name = self.get_css_class_name(ann)
            start_payload = self.get_tsawa_sp(css_class_name)
            end_payload = Tsadra_template.span_EP
        elif ann_type == LayerEnum.citation:
            css_class_name = self.get_css_class_name(ann)
            start_payload = self.get_citation_sp(css_class_name)
            end_payload = Tsadra_template.span_EP
        elif ann_type == LayerEnum.sabche:
            start_payload = Tsadra_template.sabche_SP
            end_payload = Tsadra_template.span_EP
        elif ann_type == LayerEnum.yigchung:
            start_payload = Tsadra_template.yigchung_SP
            end_payload = Tsadra_template.span_EP
        elif ann_type == LayerEnum.footnote:
            start_payload = f'<a href="#fr{ann["id"]}">{Tsadra_template.footnote_marker_SP} id="fm{ann["id"]}">'
            end_payload = Tsadra_template.footnote_EP

        start_cc, end_cc = self.__get_adapted_span(ann["span"], base_id)
        self.add_chars(base_id, start_cc, True, start_payload)
        if not only_start_ann:
            self.add_chars(base_id, end_cc, False, end_payload)

    def p_tag_adder(self, body_text):
        """Add p tag to lines where it is missing

        Args:
            body_text (str): body tag of ebook after layers are applied

        Returns:
            str: body tag with proper p tags
        """
        new_body_text = ""
        body_text = re.sub(r"\n</span>", "\n</span>\n", body_text)
        paras = body_text.split("\n")
        para_flag = False
        cur_span_payload = ""
        cur_para = ""
        for para in paras:
            if "<p" not in para:
                if re.search("<span.+?</span>", para):
                    new_body_text += f"<p>{para}</p>"
                elif "<span" in para and not para_flag:
                    cur_span_payload = re.search("<span .+>", para)[0]
                    cur_para += f"<p>{para}</span></p>"
                    para_flag = True
                elif para_flag and "</span>" not in para:
                    cur_para += (
                        f'<p>{cur_span_payload[:-2]}-middle-line">{para}</span></p>'
                    )
                elif "</span>" in para:
                    cur_para += f'<p>{cur_span_payload[:-2]}-last-line">{para}</p>'
                    new_body_text += cur_para
                    cur_para = ""
                    para_flag = False
                    cur_span_payload = ""
                else:
                    new_body_text += f"<p>{para}</p>"
            else:
                new_body_text += para
        return new_body_text

    def is_title(self, p_tag):
        """Check if a p tag is title type annotation such as title, subtitle, chapter or book-number

        Args:
            p_tag (str): p tag of a body text

        Returns:
            boolean: True if p tag is title type annotation else False
        """
        if "title" in p_tag or "chapter" in p_tag or "number" in p_tag:
            return True
        else:
            return False

    def is_sabche_only(self, p_tag):
        """Check if p tag is non inline sabche or not

        Args:
            p_tag (str): p tag of a body text

        Returns:
            boolean: True if p tag is non inline sache else False
        """
        if re.search('<p><span class="tibetan-sabche1">', p_tag):
            return True
        else:
            return False

    def rm_indentation(self, p_tag):
        """Return p tag with non indented style

        Args:
            p_tag (str): p tag of a body text

        Returns:
            str: p tag with non indented style
        """
        if self.is_sabche_only(p_tag):
            p_tag = p_tag.replace(
                '<p><span class="tibetan-sabche1">', '<p><span class="tibetan-sabche">'
            )
        if len(self.get_p_text(p_tag)) > 50:
            p_tag = re.sub("<p>", '<p class="tibetan-commentary-non-indent">', p_tag)
        else:
            p_tag = p_tag.replace("<p>", '<p class="tibetan-regular-indented">')
        return p_tag

    def add_indentation(self, p_tag):
        """Add indentation style to p tag

        Args:
            p_tag (str): p tag of boody text

        Returns:
            str: p tag with indentation style
        """
        p_tag = re.sub("<p>", '<p class="tibetan-regular-indented">', p_tag)
        return p_tag

    def is_annotated_p_tag(self, p_tag):
        """Check a p tag is annotated p tag or not

        Args:
            p_tag (str): p tag of a body text

        Returns:
            boolean: True if p tag is annotated else False
        """
        if re.search("<p><span.*?>.+?</span></p>", p_tag) or re.search(
            '<p class=".+"><span class=".+">.+?</span></p>', p_tag
        ):
            return True
        else:
            return False

    def get_p_text(self, p_tag):
        """Return text included in p tag

        Args:
            p_tag (str): p tag of body text

        Returns:
            str: text of p tag
        """
        para_tag = re.search(r"<p.*?>(<span.*?>)?(.+?)(</span>)?</p>", p_tag)
        para_text = para_tag.group(2)
        return para_text

    def get_p_tags(self, body_text):
        """Return all the p tags in body text

        Args:
            body_text (str): body text of the ebook

        Returns:
            list: p tags in body text
        """
        p_tags = re.split("(<p.*?>.+?</p>)", body_text)
        return p_tags[1::2]

    def has_yigchung(self, p_tag):
        """Check if p tag contains yigchung or not

        Args:
            p_tag (str): p tag of a body

        Returns:
            boolean: True of yigchung exist in p tag else False
        """
        if Tsadra_template.yigchung_SP in p_tag:
            return True
        else:
            return False

    def get_front_page(self):
        """Construct the front page of the book including title, subtitle, authors and credit page

        Returns:
            str: front page content
        """
        front_page = ""
        source_metadata = self.meta.get("source_metadata", {})
        credit_pg_name = source_metadata.get("credit", "")
        credit_pg_path = self.opf_path / f"assets/image/{credit_pg_name}"
        book_title_tag = f'<p class="tibetan-regular-indented">{Tsadra_template.cover_title_SP}{source_metadata.get("title", "")}</span></p>\n'
        sub_title_tag = f'<p class="tibetan-regular-indented">{Tsadra_template.sub_title_SP}{source_metadata.get("subtitle", "")}</span></p>\n'
        authors = source_metadata.get("authors", [])
        front_page += f"{book_title_tag}{sub_title_tag}"
        for author in authors:
            front_page += f"{Tsadra_template.author_SP}{author}</span></p>\n"
        if credit_pg_path.is_file():
            front_page += f'{Tsadra_template.credit_page_SP}<img src="{self.opf_path}/assets/image/{credit_pg_name}" alt="credit image not found"/></span></p>\n'
        return front_page

    def add_page_break(self, prev_p_tag, body_text):
        """Add page break before title annotation type p tag

        Args:
            prev_p_tag (str): prev tag of the title type p tag
            body_text (str): body text of ebook

        Returns:
            str: body text with page break before title type p tag
        """
        if not self.is_title(prev_p_tag) and (
            len(self.get_p_text(prev_p_tag)) > 500 or self.has_yigchung(prev_p_tag)
        ):
            new_prev_p_tag = prev_p_tag.replace(
                '<p class="tibetan-commentary-non-indent">',
                '<p class="tibetan-commentary-non-indent1">',
            )
            new_prev_p_tag = new_prev_p_tag.replace(
                '<p class="tibetan-regular-indented">',
                '<p class="tibetan-regular-indented1">',
            )
        else:
            new_prev_p_tag = prev_p_tag
        return body_text.replace(prev_p_tag, new_prev_p_tag)

    def indentation_adjustment(self, body_text):
        """Add required indentation styles to body text's p tags

        Args:
            body_text (str): body of the ebook

        Returns:
            str: body text having p tags with expected indentation styles
        """
        p_tags = self.get_p_tags(body_text)

        prev_p_tag = p_tags[0]
        body_text = prev_p_tag
        for p_tag in p_tags[1:]:
            cur_p_tag = ""
            if self.is_annotated_p_tag(prev_p_tag):
                if not self.is_title(p_tag):
                    cur_p_tag = self.rm_indentation(p_tag)
                else:
                    body_text = self.add_page_break(prev_p_tag, body_text)
                    cur_p_tag = self.add_indentation(p_tag)
            else:
                if self.is_annotated_p_tag(p_tag):
                    if not self.is_title(p_tag):
                        cur_p_tag = self.rm_indentation(p_tag)
                    else:
                        body_text = self.add_page_break(prev_p_tag, body_text)
                        cur_p_tag = self.add_indentation(p_tag)
                else:
                    if (
                        len(self.get_p_text(prev_p_tag)) < 50
                        and len(self.get_p_text(p_tag)) > 150
                    ):
                        cur_p_tag = p_tag.replace(
                            "<p>", '<p class="tibetan-commentary-non-indent">'
                        )
                    else:
                        cur_p_tag = self.add_indentation(p_tag)
            body_text += cur_p_tag
            prev_p_tag = cur_p_tag
        return body_text

    def get_footnote_references(self, footnotes):
        """Generate footnote references using footnote layer

        Args:
            footnotes (dict): footnote annotation information

        Returns:
            str: footnote references p tags
        """
        footnote_references = ""
        p_tag = '<p class = "tibetan-commentary-non-indent">'
        for footnote_id, footnote in footnotes.items():
            footnote_references += f'{p_tag}<a href="#fm{footnote_id}">{Tsadra_template.footnote_reference_SP} id="fr{footnote_id}">{footnote["footnote_ref"]}</span></a></p>'
        return footnote_references

    def set_toc_level(self, toc_levels, serialized_html):
        """Return toc outline by checking presences of title type annotations and sabche

        Args:
            toc_levels (dict): toc level outline given by user
            serialized_html (str): final serialize html of the opf

        Returns:
            dict: toc outline created using existing title type annotations or sabche
        """
        new_toc_levels = {}
        walker = 1
        for level, annotation_name in toc_levels.items():
            annotation_xpath = Tsadra_template.toc_xpaths.get(annotation_name, "")
            if annotation_xpath and annotation_name in serialized_html:
                new_toc_levels[walker] = annotation_xpath
                walker += 1
        return new_toc_levels

    def get_new_metadata(self, soup, meta_data):
        """Update meta data of opf

        Args:
            soup (bs4-tag): beautiful soup tag of whole opf
            meta_data (bs4-tag): beautiful soup tag of metadata

        Returns:
            bs4-tag: new metadata
        """
        ibook_meta = soup.new_tag("meta")
        ibook_meta.attrs["property"] = "ibooks:specified-fonts"
        ibook_meta.append("true")
        meta_data.append(ibook_meta)
        return meta_data

    def get_new_manifest(self, soup, manifest):
        """Update manifest of opf

        Args:
            soup (bs4-tag): beautiful soup tag of whole opf
            manifest (bs4-tag): beautiful soup tag of manifest

        Returns:
            bs4-tag: new manifest
        """
        new_item = soup.new_tag("item")
        new_item.attrs["href"] = "font/MonlamUniOuChan2.ttf"
        new_item.attrs["id"] = "MonlamUniOuChan2.ttf"
        new_item.attrs["media-type"] = "application/x-font-ttf"
        manifest.append(new_item)
        return manifest

    def get_new_opf(self, opf):
        """Updating opf content of epub with ibook specifications

        Args:
            opf (str): opf content of epub produced by calibre

        Returns:
            str: new opf content
        """
        soup = BeautifulSoup(opf, "html.parser")
        soup.package.attrs[
            "prefix"
        ] = "ibooks: http://vocabulary.itunes.apple.com/rdf/ibooks/vocabulary-extensions-1.0/"
        old_meta = soup.find("metadata")
        old_manifest = soup.find("manifest")
        new_meta = self.get_new_metadata(soup, old_meta)
        new_manifest = self.get_new_manifest(soup, old_manifest)
        soup.metadata.replaceWith(new_meta)
        soup.manifest.replaceWith(new_manifest)
        return str(soup)

    def embed_ibook_specific_font(self, epub_path):
        """Include ibook specification for proper font embedding and recompiling the ebook.

        Args:
            epub_path (path_obj): path of epub produced by calibre
        """
        pecha_id = epub_path.stem
        # rename to zip
        zip_path = epub_path.parent / f"{pecha_id}.zip"
        os.rename(str(epub_path), str(zip_path))
        # epub_path.rename(epub_path.with_suffix('.zip'))
        epub_folder = str(epub_path.parent / f"{pecha_id}")
        with zipfile.ZipFile(zip_path, "r") as zip_ref:
            zip_ref.extractall(epub_folder)
        print("INFO: Unzip epub..")
        zip_path.unlink()
        opf_path = Path(epub_folder) / "content.opf"
        opf_content = opf_path.read_text(encoding="utf-8")
        new_opf_content = self.get_new_opf(opf_content)
        opf_path.write_text(new_opf_content, encoding="utf-8")
        print("INFO: OPF content updated..")
        new_zip_path = shutil.make_archive(epub_folder, "zip", epub_folder)
        os.system(f"rm -r {epub_folder}")
        print("INFO: file zipped..")
        new_zip_path = Path(new_zip_path)
        new_epub_path = f"{epub_folder}.epub"
        os.rename(str(new_zip_path), new_epub_path)
        # new_zip_path.rename(new_zip_path.with_suffix('.epub'))
        print("INFO: Epub ready...")

    def get_serialized_html(self, result, base_id, pecha_title):
        """Serialize html are return using annotated text

        Args:
            result (str): tsadra epub annotations applied text
            base_id (str): volume id
            pecha_title (str): pecha title

        Returns:
            str: annotated applied text in html
        """
        result = f"{self.get_front_page()}{result}"
        footnote_ref_tag = ""
        if "Footnote" in self.layers:
            footnote_fn = self.opf_path / "layers" / base_id / "Footnote.yml"
            footnote_layer = load_yaml(footnote_fn)
            footnote_ref_tag = self.get_footnote_references(
                footnote_layer["annotations"]
            )
        result = self.p_tag_adder(result)
        result = self.indentation_adjustment(result)
        serialized_html = (
            f"<html>\n<head>\n\t<title>{pecha_title}</title>\n</head>\n<body>\n"
        )
        serialized_html += f"{result}{footnote_ref_tag}</body>\n</html>"
        return serialized_html

    def serialize(
        self, output_path="./output/epub_output", toc_levels={}, css_template=None
    ) -> Path:
        """
        This module serialize .opf file to other format such as .epub etc. In case of epub,
        we are using calibre ebook-convert command to do the conversion by passing our custom css template
        and embedding our custom font. The converted output will be then saved in current directory
        as {pecha_id}.epub.

        Args:
            toc_levels (dict): table of content level
            output_path (str): output path where epub needs to be saved

        Returns:
            str: output path

        """
        output_path = Path(output_path)
        out_html_fn = f"{self.meta['id']}.html"
        pecha_title = self.meta["source_metadata"].get("title", "")
        cover_image = self.meta["source_metadata"].get("cover", "")

        self.apply_layers()
        self.layers = [layer for layer in self.layers if layer != "Pagination"]

        results = self.get_result()
        for base_id, result in results.items():
            serialized_html = self.get_serialized_html(result, base_id, pecha_title)
            Path(out_html_fn).write_text(serialized_html, encoding='utf-8')
            # Downloading css template file from ebook template repo and saving it
            if not css_template:
                css_template = requests.get(
                    "https://github.com/OpenPecha-Data/ebook-template/blob/master/tsadra_template.css"
                )
                css_template = str(css_template.content)
            Path("template.css").write_text(css_template)
            # Running ebook-convert command to convert html file to .epub (From calibre)
            # XPath expression to detect chapter titles.
            if not toc_levels:
                toc_levels = Tsadra_template.toc_levels
            toc_levels = self.set_toc_level(toc_levels, serialized_html)
            level1_toc_Xpath = toc_levels.get(1, "")
            level2_toc_Xpath = toc_levels.get(2, "")
            level3_toc_Xpath = toc_levels.get(3, "")

            cover_path = self.opf_path / f"assets/image/{cover_image}"
            out_epub_fn = output_path / f"{self.meta['id']}.epub"
            font_family = "Monlam Uni Ouchan2"
            if cover_path.is_file():
                os.system(
                    f'ebook-convert {out_html_fn} {out_epub_fn} --extra-css=./template.css --embed-font-family="{font_family}" --page-breaks-before="{Tsadra_template.book_title_Xpath}" --cover={cover_path} --flow-size=0 --level1-toc="{level1_toc_Xpath}" --level2-toc="{level2_toc_Xpath}" --level3-toc="{level3_toc_Xpath}" --use-auto-toc --disable-font-rescaling'
                )
            else:
                os.system(
                    f'ebook-convert {out_html_fn} {out_epub_fn} --extra-css=./template.css --embed-font-family="{font_family}" --page-breaks-before="{Tsadra_template.book_title_Xpath}" --flow-size=0 --level1-toc="{level1_toc_Xpath}" --level2-toc="{level2_toc_Xpath}" --level3-toc="{level3_toc_Xpath}" --use-auto-toc --disable-font-rescaling'
                )
            # Removing html file and template file
            os.system(f"rm {out_html_fn}")
            os.system("rm template.css")
            if out_epub_fn.is_file():
                self.embed_ibook_specific_font(out_epub_fn)
            return out_epub_fn



"""This module is to create pecha opf and alignment opa
Avialable functions:
    create_alignment_from_source_text: It will create pecha opf and alignment opa from a source text.
    create_alignment_from_tmx: It will create source and target opf and alignment opa from tmx path.
"""

import os
import shutil
from pathlib import Path
from uuid import uuid4
import datetime
from datetime import timezone
from enum import Enum

from openpecha import config, github_utils
from openpecha.core.annotation import AnnBase, Span
from openpecha.core.layer import Layer, LayerEnum
from openpecha.core.pecha import OpenPechaGitRepo
from openpecha.utils import load_yaml, dump_yaml
from openpecha.core.ids import get_base_id, get_initial_pecha_id
from openpecha.core.metadata import InitialPechaMetadata, InitialCreationType

from ..parsers.tmx import parse_tmx
from ..segmenters.sentence import get_sentence_segments
from . import TMXAlignment


def get_segment_annotation(segment, char_walker):
    seg_len = len(segment) - 1
    segment_annotation = {
        uuid4().hex: AnnBase(span=Span(start=char_walker, end=char_walker + seg_len))
    }
    return segment_annotation, (char_walker + seg_len)


def get_segment_layer(text):
    segment_annotations = {}
    char_walker = 0
    segments = text.splitlines()
    for segment in segments:
        if segment:
            segment_annotation, end = get_segment_annotation(segment, char_walker)
            segment_annotations.update(segment_annotation)
        char_walker = end + 2

    segment_layer = Layer(
        annotation_type=LayerEnum.segment, annotations=segment_annotations
    )
    return segment_layer


def get_base_text(text):
    final_base = ""
    segments = text.splitlines()
    for segment in segments:
        final_base += segment + "\n"
    return final_base


def get_metadata(pecha_id: str = None, language: str = None, source_metadata: dict = None) -> InitialPechaMetadata:
        copyright = {}
        license = None
        parser_link = None

        metadata = InitialPechaMetadata(
            id=pecha_id,
            source='',
            initial_creation_type=InitialCreationType.tmx,
            imported=datetime.datetime.now(timezone.utc),
            last_modified=datetime.datetime.now(timezone.utc),
            parser=parser_link,
            copyright=copyright,
            license=license,
            source_metadata=source_metadata,
            default_language=language
        )
        return metadata


def create_readme(pecha_path):
    pecha_id = pecha_path.stem
    meta_yml = load_yaml((pecha_path / f"{pecha_id}.opf" / "meta.yml"))
    pecha = f"|Pecha id | {pecha_id}"
    Table = "| --- | --- "
    Title = f"|Title | {meta_yml.get('source_metadata', {}).get('title', None)} "
    lang = f"|Language | {meta_yml.get('default_language', None)}"
    Creator = f"|Initial creation type | { meta_yml.get('initial_creation_type', None)}"
    readme = f"{pecha}\n{Table}\n{Title}\n{lang}\n{Creator}"
    return readme


def create_opf(text, lang=None, source_metadata=None, new=False):
    if new:
        text = get_sentence_segments(text)
    pecha_id = get_initial_pecha_id()
    opf_path = config.PECHAS_PATH / pecha_id / f"{pecha_id}.opf"
    opf_path.mkdir(exist_ok=True, parents=True)
    pecha = OpenPechaGitRepo(path=opf_path)
    base_id = get_base_id()
    layers = {f"{base_id}": {LayerEnum.segment: get_segment_layer(text)}}
    base_text = get_base_text(text)
    bases = {f"{base_id}": base_text}
    metadata = get_metadata(pecha_id, lang, source_metadata)
    pecha.layers = layers
    pecha.bases = bases
    pecha._meta = metadata
    
    pecha._meta.bases = {
        base_id:
            {   
                "source_metadata": None,
                "order": 1,
                "base_file": f"{base_id}.txt",
                "statistics": None
                }
            }
    pecha.save_base()
    pecha.save_layers()
    pecha.save_meta()
    readme = create_readme(pecha.pecha_path)
    (pecha.pecha_path / "readme.md").write_text(readme, encoding="utf-8")
    return pecha


def publish_alignment(pecha_path, asset_path):
    asset_paths = []
    github_utils.github_publish(
        pecha_path,
        message="initial commit",
        not_includes=[],
        layers=[],
        org=os.environ.get("OPENPECHA_DATA_GITHUB_ORG"),
        token=os.environ.get("GITHUB_TOKEN"),
        description=f"title: {asset_path.stem}"
        )
    if asset_path:
        repo_name = pecha_path.stem
        asset_name = asset_path.stem
        shutil.make_archive(asset_path.parent / asset_name, "zip", asset_path)
        asset_paths.append(f"{asset_path.parent / asset_name}.zip")
        github_utils.create_release(
            repo_name,
            prerelease=False,
            asset_paths=asset_paths, 
            org=os.environ.get("OPENPECHA_DATA_GITHUB_ORG"),
            token=os.environ.get("GITHUB_TOKEN")
        )

def create_opf_from_tmx(tmx_path):
    title = tmx_path.stem

    src_text, tar_text, source_metadata = parse_tmx(tmx_path)

    src_lang = source_metadata.get("srclang", "")
    tar_lang = source_metadata.get("adminlang", "")

    source_metadata.update({'title':title})
    
    source_pecha = create_opf(src_text, src_lang, source_metadata)
    target_pecha = create_opf(tar_text, tar_lang, source_metadata)

    return source_pecha, target_pecha, source_metadata


def create_alignment_from_source_text(text_path, lang, source_metadata=None, publish=True):
    text = Path(text_path).read_text(encoding="utf-8")
    pecha = create_opf(text, lang, source_metadata, new=True)
    obj = TMXAlignment()
    alignment_path = obj.create_alignment_repo(
        source_pecha=pecha,
        target_pecha=None,
        title=source_metadata['title'],
        source_metadata=None,
    )
    if publish:
        pecha.publish(asset_path=text_path.parent, asset_name=text_path.stem)
        publish_alignment(alignment_path, text_path)
    return alignment_path


def create_alignment(source_pecha, target_pecha, title, source_metadata):
    obj = TMXAlignment()
    alignment_path = obj.create_alignment_repo(
        source_pecha, target_pecha, title, source_metadata
    )
    return alignment_path


def create_alignment_from_tmx(tmx_path, publish=True):
    title = tmx_path.stem
    source_pecha, target_pecha, source_metadata = create_opf_from_tmx(
        tmx_path
    )
    alignment_path = create_alignment(
        source_pecha, target_pecha, title, source_metadata
    )

    if publish:
        source_pecha.publish(asset_path=tmx_path.parent, asset_name=tmx_path.stem)
        target_pecha.publish(asset_path=tmx_path.parent, asset_name=tmx_path.stem)
        publish_alignment(alignment_path, asset_path=tmx_path.parent)
        
    return alignment_path



import gzip
import json
import logging
import statistics
from fontTools import unicodedata

from openpecha.formatters.ocr.ocr import OCRFileProvider, OCRFormatter, BBox, UNICODE_CHARCAT_FOR_WIDTH
from openpecha.buda.api import get_buda_scan_info, get_image_list, image_group_to_folder_name

class GoogleVisionBDRCFileProvider(OCRFileProvider):
    def __init__(self, bdrc_scan_id, ocr_import_info, ocr_disk_path=None, mode="local"):
        # ocr_base_path should be the output/ folder in the case of BDRC OCR files
        self.ocr_import_info = ocr_import_info
        self.ocr_disk_path = ocr_disk_path
        self.bdrc_scan_id = bdrc_scan_id
        self.mode = mode

    def get_image_list(self, image_group_id):
        buda_il = get_image_list(self.bdrc_scan_id, image_group_id)
        # format should be a list of image_id (/ file names)
        return map(lambda ii: ii["filename"], buda_il)

    def get_source_info(self):
        return get_buda_scan_info(self.bdrc_scan_id)

    def get_image_data(self, image_group_id, image_id):
        # TODO: implement the following modes:
        #  - "s3" (just read images from s3)
        #  - "s3-localcache" (cache s3 files on disk)
        # TODO: handle case where only zip archives are present on s3, one per volume.
        #       This should be indicated in self.ocr_import_info["ocr_info"]
        vol_folder = image_group_to_folder_name(self.bdrc_scan_id, image_group_id)
        expected_ocr_filename = image_id[:image_id.rfind('.')]+".json.gz"
        expected_ocr_path = self.ocr_disk_path / vol_folder / expected_ocr_filename
        ocr_object = None
        try:
            ocr_object = json.load(gzip.open(str(expected_ocr_path), "rb"))
        except:
            logging.exception("could not read "+str(expected_ocr_path))
        return ocr_object

class GoogleVisionFormatter(OCRFormatter):
    """
    OpenPecha Formatter for Google OCR JSON output of scanned pecha.
    """

    def __init__(self, output_path=None, metadata=None):
        super().__init__(output_path, metadata)

    def has_space_attached(self, symbol):
        """Checks if symbol has space followed by it or not

        Args:
            symbol (dict): symbol info

        Returns:
            boolean: True if the symbol has space followed by it
        """
        if ('property' in symbol and 
                'detectedBreak' in symbol['property'] and 
                'type' in symbol['property']['detectedBreak'] and 
                symbol['property']['detectedBreak']['type'] == "SPACE"):
            return True
        return False

    def get_language_code_from_gv_poly(self, gv_poly):
        lang = ""
        properties = gv_poly.get("property", {})
        if properties:
            languages = properties.get("detectedLanguages", [])
            if languages:
                lang = languages[0]['languageCode']
        if lang == "" or lang == "und":
            # this is not always true but replacing it with None is worse
            # with our current data
            return self.default_language
        if lang in ["bo", "en", "zh"]:
            return lang
        if lang == "dz":
            return "bo"
        # English is a kind of default for our purpose
        return "en"

    @staticmethod
    def get_bboxinfo_from_vertices(vertices):
        """
        Vertices do not always have dots in the same order. The current hypothesis
        is that their order represents the rotation of characters detected by
        the OCR.

        This is not documented on

        https://cloud.google.com/vision/docs/reference/rest/v1/projects.locations.products.referenceImages#Vertex

        though, so use the angle value with caution.
        """
        if len(vertices) == 0:
            return None
        idx_smallest = -1
        smallest_x = -1
        smallest_y = -1
        largest_x = -1
        largest_y = -1
        for idx, v in enumerate(vertices):
            if "x" not in v or "y" not in v:
                continue
            smallest_x = v["x"] if smallest_x == -1 else min(v["x"], smallest_x)
            smallest_y = v["y"] if smallest_y == -1 else min(v["y"], smallest_y)
            largest_x = max(v["x"], largest_x)
            largest_y = max(v["y"], largest_y)
            # here we have to account for cases where the 4 dots don't form a rectangle
            # because coordinates are shifted by 1, see test_bbox_info for some example
            if abs(v["x"] - smallest_x) < 3 and abs(v["y"] - smallest_y) < 3:
                idx_smallest = idx
        if smallest_x == -1 or smallest_y == -1 or largest_y == -1 or largest_x == -1:
            return None
        angle = None
        if len(vertices) == 4 and idx_smallest != -1:
            angle = 0
            if idx_smallest == 1:
                angle = 270
            if idx_smallest == 2:
                angle = 180
            if idx_smallest == 3:
                angle = 90
        return [smallest_x, largest_x, smallest_y, largest_y, angle]

    def dict_to_bbox(self, word):
        """Convert bounding bbox to BBox object

        Args:
            word (dict): bounding gv_poly of a word infos

        Returns:
            obj: BBox object of bounding bbox
        """
        confidence = word.get('confidence')
        if 'boundingBox' not in word or 'vertices' not in word['boundingBox']:
            return None
        vertices = word['boundingBox']['vertices']
        bboxinfo = GoogleVisionFormatter.get_bboxinfo_from_vertices(vertices)
        if bboxinfo == None:
            return None
        if self.remove_rotated_boxes and bboxinfo[4] > 0:
            return None
        return BBox(bboxinfo[0], bboxinfo[1], bboxinfo[2], bboxinfo[3], bboxinfo[4], 
            confidence=confidence)

    @staticmethod
    def get_width_of_vertices(vertices):
        if len(vertices) < 4:
            return None
        # oddly enough, sometimes Google returns a vertex with x=-1...
        smallest_x = None
        largest_x = -1
        for v in vertices:
            if "x" not in v or "y" not in v:
                continue
            smallest_x = v["x"] if smallest_x == None else min(v["x"], smallest_x)
            largest_x = max(v["x"], largest_x)
        if smallest_x == None:
            return None
        return largest_x - smallest_x

    def get_char_base_bboxes_and_avg_width(self, response):
        """Return bounding bboxs in page response

        Args:
            response (dict): google ocr output of a page

        Returns:
            list: list of BBox object which saves required info of a bounding bbox
        """
        bboxes = []
        widths = []
        for page in response['fullTextAnnotation']['pages']:
            for block in page['blocks']:
                for paragraph in block['paragraphs']:
                    for word in paragraph['words']:
                        bbox = self.dict_to_bbox(word)
                        if bbox is None:
                            # case where we ignore the bbox for some reason
                            # for instance rotated text
                            continue
                        cur_word = ""
                        for symbol in word['symbols']:
                            symbolunicat = unicodedata.category(symbol['text'][0])
                            if symbolunicat in UNICODE_CHARCAT_FOR_WIDTH:
                                vertices = symbol['boundingBox']['vertices']
                                width = GoogleVisionFormatter.get_width_of_vertices(vertices)
                                if width is not None and width > 0:
                                    widths.append(width)
                            cur_word += symbol['text']
                            if self.has_space_attached(symbol):
                                cur_word += " "
                        if cur_word:
                            bbox.text = cur_word
                            # the language returned by Google OCR is not particularly helpful
                            # language = self.get_language_code_from_gv_poly(word)
                            # instead we use our custom detection system
                            bbox.language = self.get_main_language_code(cur_word)
                            bboxes.append(bbox)
        avg_width = statistics.mean(widths) if widths else None
        logging.debug("average char width: %f", avg_width)
        return bboxes, avg_width

    def get_bboxes_for_page(self, image_group_id, image_filename):
        ocr_object = self.data_provider.get_image_data(image_group_id, image_filename)
        try:
            page_content = ocr_object["textAnnotations"][0]["description"]
        except Exception:
            logging.error("OCR page is empty (no textAnnotations[0]/description)")
            return None, 0
        return self.get_char_base_bboxes_and_avg_width(ocr_object)



from pydoc import pager
import re
import os
from enum import Enum
from pathlib import Path

import json
import logging
import datetime
import statistics
from zipfile import ZipFile
from datetime import timezone
from bs4 import BeautifulSoup

from openpecha.formatters.ocr.ocr import OCRFormatter, BBox

# from openpecha.core.annotation import Page, Span
from openpecha.utils import load_yaml, dump_yaml, gzip_str
# from openpecha.formatters.google_vision import GoogleVisionFormatter, GoogleVisionBDRCFileProvider
from openpecha.buda.api import get_buda_scan_info, get_image_list, image_group_to_folder_name


class BDRCGBFileProvider():
    def __init__(self, bdrc_scan_id, buda_data, ocr_import_info, ocr_disk_path):
        self.ocr_import_info = ocr_import_info
        # disk path is the path to a directory that contains "info" and "output" subfolders
        self.ocr_disk_path = ocr_disk_path
        self.bdrc_scan_id = bdrc_scan_id
        self.buda_data = buda_data
        self.images_info = {}
        self.cur_zip = None
        self.cur_image_group_id = None
    

    def get_image_list(self, image_group_id):
        """call get_images_info and return the image_list of the image_group_id

        Args:
            image_group_id (str): image_group_id of the volume
        """
        self.get_images_info(image_group_id)
        buda_il = get_image_list(self.bdrc_scan_id, image_group_id)
        # format should be a list of image_id (/ file names)
        return map(lambda ii: ii["filename"], buda_il)
    
    
    def get_images_info(self, image_group_id):
        """load page_filename with image_filename mapping to images.info

        Args:
            image_group_id (str): image_group_id of the volume
        """
        vol_folder = image_group_to_folder_name(self.bdrc_scan_id, image_group_id)
        vol_path = Path(f"{self.ocr_disk_path}") / "info" / vol_folder
        if os.path.isdir(vol_path):
            image_info_path = Path(vol_path) / "gb-bdrc-map.json"
            self.images_info = load_yaml(image_info_path)
        else:
            return

    def get_source_info(self):
        return self.buda_data
    
    def get_hocr_filename(self, image_id):
        """ return the filename of the image_id from the images_info dict
        """
        for filename, img_ref in self.images_info.items():
            img_id = img_ref
            if img_id == image_id:
                return filename

    def get_image_group_data(self, image_group_id):
        """unzip the html.zip of the image_group_id and return it

        Args:
            image_group_id (str): image_group_id of the volume

        Returns:
            self.cur_zip: unzip content of the html.zip
        """
        if image_group_id == self.cur_image_group_id and self.cur_zip is not None:
            return self.cur_zip
        vol_folder = image_group_to_folder_name(self.bdrc_scan_id, image_group_id)
        zip_path = Path(f"{self.ocr_disk_path}") / "output" / vol_folder / "html.zip"
        self.cur_zip = ZipFile(zip_path)
        return self.cur_zip
    
    def get_image_data(self, image_group_id, image_filename):
        """get hocr_filename using image_filename,
           use volume folder and hocr_filename to read the hocr_html file for the image_filename or the page

        Args:
            image_group_id (str): image_group_id of the volume

        Returns:
            hocr_html: html file of the image_filename
        """
        hocr_filename = self.get_hocr_filename(image_filename)+".html"
        zf = self.get_image_group_data(image_group_id)
        try:
            for filename in zf.filelist:
                if filename.filename.split("/")[-1] == hocr_filename:
                    with zf.open(filename.filename) as hocr_file:
                        return hocr_file.read()
        except:
            return 

class HOCRIAFileProvider():
    def __init__(self, bdrc_scan_id, bdrc_image_list_path, buda_data, ocr_import_info, ocr_disk_path=None):
        self.ocr_import_info = ocr_import_info
        self.ocr_disk_path = ocr_disk_path
        self.bdrc_scan_id = bdrc_scan_id
        self.buda_data = buda_data
        self.bdrc_image_list_path = bdrc_image_list_path
        self.images_info = {}


    def get_image_list(self, image_group_id):
        """call get_images_info and return the image_list of the image_group_id

        Args:
            image_group_id (str): image_group_id of the volume
        """
        self.get_images_info(image_group_id)
        buda_il = get_image_list(self.bdrc_scan_id, image_group_id)
        # format should be a list of image_id (/ file names)
        return map(lambda ii: ii["filename"], buda_il)
    
    
    def get_images_info(self, image_group_id):
        """map the image_filename to page_info from the hocr_html and populate the images_info

        Args:
            image_group_id (str): image_group_id of the volume
        """
        curr_image = {}
        image_list = get_image_list(self.bdrc_scan_id, image_group_id)
        image_group_hocr = self.get_image_group_hocr(image_group_id)
        if image_group_hocr :
            hocr_html = BeautifulSoup(image_group_hocr, 'html.parser')
            pages_hocr = hocr_html.find_all("div", {"class": "ocr_page"})
            for image_number, image_filename in enumerate (image_list):
                filename = image_filename['filename']
                for page_hocr in pages_hocr:
                    if int(page_hocr['id'][5:]) == image_number:
                        curr_image[filename] = {
                            'page_info': page_hocr
                        }
                        self.images_info.update(curr_image)
                        curr_image = {}
                
    
    def get_image_group_hocr(self, image_group_id):
        """use volume number and bdrc scan id to read the hocr_html file for the volume

        Args:
            image_group_id (str): image_group_id of the volume

        Returns:
            hocr_html: html file of the volume
        """
        vol_num = self.buda_data['image_groups'][image_group_id]['volume_number']
        image_group_hocr_path = Path(f"{self.bdrc_image_list_path}") / f"bdrc-{self.bdrc_scan_id}-{vol_num}_hocr.html"
        try:
            hocr_html = image_group_hocr_path.read_text(encoding='utf-8')
            return hocr_html
        except:
            return
    
    def get_source_info(self):
        return self.buda_data
        
    
    def get_image_data(self, image_group_id, image_filename):
        """use images_info to return the page_info of the image_filename

        Args:
            image_group_id (str): image_group_id of the volume
        """
        try:
            page_hocr = self.images_info[image_filename]['page_info']
            return page_hocr
        except:
            return
    
class HOCRFormatter(OCRFormatter):
    """
    OpenPecha Formatter for Google OCR HOCR output of scanned pecha.
    """
    def __init__(self, mode=None, output_path=None, metadata=None):
        super().__init__(output_path, metadata)
        self.mode = mode
        self.word_span = 0
    
    def get_word_text_with_space(self, line_text, word_box):
        """check for space after word_text using line_text, add space to word_text if found 
        """
        text = word_box.text
        self.word_span += len(text)
        if len(line_text) == self.word_span:
            return text
        else:
            next_character = line_text[self.word_span]
            if next_character == " ":
                self.word_span += 1
                text = text + " "
        return text
        
    def parse_box(self, line_box, word_box):
        """parse the word_box to create bbox for the text in word_box.

       Args:
            line_text (_type_): line from the page html 
            word_box (_type_): word from the above line

        Returns:
            box : bbox for text in word_box with vertices, confidence, language 
        """
        line_text = line_box.text
        if not word_box.has_attr('title'):
            return None
        boxinfos = word_box['title'].split(';')
        coords = None
        angle = None
        confidence = None
        for boxinfo in boxinfos:
            boxinfo_parts = boxinfo.strip().split(" ")
            if boxinfo_parts[0] == "bbox":
                # in HOCR's, bbox order is x0, y0, x1, y1
                coords = [
                    int(boxinfo_parts[1]), 
                    int(boxinfo_parts[2]),
                    int(boxinfo_parts[3]),
                    int(boxinfo_parts[4])
                    ]
            elif boxinfo_parts[0] == "textangle":
                # angle is indicated counter-clockwise in hocr so
                # we need to convert it to our internal value system:
                angle = int(boxinfo_parts[1])
                if textangle != 0:
                    angle = 360 - angle
            elif boxinfo_parts[0] == "x_wconf":
                confidence = float(boxinfo_parts[1]) / 100.0
        if coords is None:
            return None
        if self.remove_rotated_boxes and angle is not None and angle > 0:
            return None
        language = self.get_main_language_code(word_box.text)
        text = self.get_word_text_with_space(line_text, word_box)
        # but we initialize as x1, x2, y1, y2
        box = BBox(coords[0], coords[2], coords[1], coords[3],
            angle = angle,
            text=text,
            confidence=confidence,
            language=language
        )
        return box
    
    def get_boxes(self, hocr_page_html):
        """parse the hocr_page_html for one html file per page for bboxes of the page

        Args:
            hocr_page_html (html): html of the page

        Returns:
            boxes: list of bbox dict
        """
        bboxes = []
        hocr_html = BeautifulSoup(hocr_page_html, 'html.parser')
        line_boxes = hocr_html.find_all("span", {"class": "ocr_line"})
        for line_box in line_boxes:
            self.word_span = 0
            word_boxes = line_box.find_all("span", {"class": "ocrx_word"})
            for word_box in word_boxes:
                bbox = self.parse_box(line_box,word_box)
                if bbox is not None:
                    bboxes.append(bbox)
        return bboxes
    
    def get_boxes_for_IA(self, page_html):
        """parse the page_html for one html file per volume for bboxes of the page

        Args:
            hocr_page_html (html): html of the page

        Returns:
            boxes: list of bbox dict
        """
        bboxes = []
        paragraphs_html = page_html.find_all("p", {"class": "ocr_par"})
        for paragraph_html in paragraphs_html:
            line_boxes = paragraph_html.find_all("span", {"class": "ocr_line"})
            for line_box in line_boxes:
                self.word_span = 0
                word_boxes = line_box.find_all("span", {"class": "ocrx_word"})
                for word_box in word_boxes:
                    bbox = self.parse_box(line_box,word_box)
                    if bbox is not None:
                        bboxes.append(bbox)
        return bboxes


    def get_bboxes_for_page(self, image_group_id, image_filename):
        bboxes = []
        hocr_page_html = self.data_provider.get_image_data(image_group_id, image_filename)
        if hocr_page_html:
            if self.mode == "IA":
                bboxes = self.get_boxes_for_IA(hocr_page_html)
            else:
                bboxes = self.get_boxes(hocr_page_html)
        return bboxes, None

    


from pathlib import Path
from openpecha.core.layer import LayerEnum
from openpecha.serializers.serialize import Serialize
from openpecha.utils import load_yaml
import mimetypes
import os


def micro_to_milliseconds(microseconds):
    return microseconds // 1000


class ElanSerializer(Serialize):
    """Elan serializer class to get .eaf file."""

    def __init__(
        self, opf_path, text_id=None, base_ids=None, layers=None, index_layer=None
    ):
        super().__init__(opf_path, text_id, base_ids, layers, index_layer)
        self.annotation_sequence = {}
        self.time_order = {}

    def apply_annotation(self, base_id, annotation, uuid2localid):

        ann_type = LayerEnum(annotation["type"])
        if ann_type == LayerEnum.transcription_time_span:
            if base_id not in self.annotation_sequence:
                self.annotation_sequence[base_id] = 1
            start_payload = """        <ANNOTATION>
            <ALIGNABLE_ANNOTATION ANNOTATION_ID="a{seq}"
                TIME_SLOT_REF1="ts{slot1}" TIME_SLOT_REF2="ts{slot2}">
                <ANNOTATION_VALUE>""".format(
                seq=self.annotation_sequence[base_id],
                slot1=self.annotation_sequence[base_id] * 2 - 1,
                slot2=self.annotation_sequence[base_id] * 2,
            )
            end_payload = """</ANNOTATION_VALUE>
            </ALIGNABLE_ANNOTATION>
        </ANNOTATION>"""
            self.add_chars(base_id, annotation["span"]["start"], True, start_payload)
            self.add_chars(base_id, annotation["span"]["end"], True, end_payload)
            if base_id not in self.time_order:
                self.time_order[
                    base_id
                ] = f"""        <TIME_SLOT TIME_SLOT_ID="ts{self.annotation_sequence[base_id] * 2 - 1}" TIME_VALUE="{micro_to_milliseconds(annotation["time_span"]["start"])}"/>
        <TIME_SLOT TIME_SLOT_ID="ts{self.annotation_sequence[base_id] * 2}" TIME_VALUE="{micro_to_milliseconds(annotation["time_span"]["end"])}"/>"""
            else:
                self.time_order[
                    base_id
                ] += f"""
        <TIME_SLOT TIME_SLOT_ID="ts{self.annotation_sequence[base_id] * 2 - 1}" TIME_VALUE="{micro_to_milliseconds(annotation["time_span"]["start"])}"/>
        <TIME_SLOT TIME_SLOT_ID="ts{self.annotation_sequence[base_id] * 2}" TIME_VALUE="{micro_to_milliseconds(annotation["time_span"]["end"])}"/>"""
            self.annotation_sequence[base_id] += 1

    def get_elan(self, base_id, result, media_url):
        media_url_path = Path(media_url)
        mime_type = (
            mimetypes.MimeTypes().guess_type(media_url_path.name)[0] or "audio/*"
        )
        return f"""<?xml version="1.0" encoding="UTF-8"?>
<ANNOTATION_DOCUMENT AUTHOR="" DATE="2023-03-15T20:50:48+05:30"
    FORMAT="3.0" VERSION="3.0"
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:noNamespaceSchemaLocation="http://www.mpi.nl/tools/elan/EAFv3.0.xsd">
    <HEADER MEDIA_FILE="" TIME_UNITS="milliseconds">
        <MEDIA_DESCRIPTOR
            MEDIA_URL="{media_url}"
            MIME_TYPE="{mime_type}" RELATIVE_MEDIA_URL=".{os.sep}{media_url_path.name}"/>
        <PROPERTY NAME="URN">urn:nl-mpi-tools-elan-eaf:ea97e0c9-0c39-48c0-be65-ae73b74d1be1</PROPERTY>
        <PROPERTY NAME="lastUsedAnnotationId">4</PROPERTY>
    </HEADER>
    <TIME_ORDER>
{self.time_order[base_id]}
    </TIME_ORDER>
    <TIER LINGUISTIC_TYPE_REF="default-lt" TIER_ID="default">
{result}    </TIER>
    <LINGUISTIC_TYPE GRAPHIC_REFERENCES="false"
        LINGUISTIC_TYPE_ID="default-lt" TIME_ALIGNABLE="true"/>
    <CONSTRAINT
        DESCRIPTION="Time subdivision of parent annotation's time interval, no time gaps allowed within this interval" STEREOTYPE="Time_Subdivision"/>
    <CONSTRAINT
        DESCRIPTION="Symbolic subdivision of a parent annotation. Annotations refering to the same parent are ordered" STEREOTYPE="Symbolic_Subdivision"/>
    <CONSTRAINT DESCRIPTION="1-1 association with a parent annotation" STEREOTYPE="Symbolic_Association"/>
    <CONSTRAINT
        DESCRIPTION="Time alignable annotations within the parent annotation's time interval, gaps are allowed" STEREOTYPE="Included_In"/>
</ANNOTATION_DOCUMENT>
"""

    def serialize(self):
        self.apply_layers()
        results = self.get_result()
        for base_id, result in results.items():
            transcription_time_span_layer_file = (
                self.opf_path / "layers" / base_id / "TranscriptionTimeSpan.yml"
            )
            transcription_time_span_layer = load_yaml(
                transcription_time_span_layer_file
            )
            elan = self.get_elan(
                base_id, result, transcription_time_span_layer["media_url"]
            )

            yield base_id, elan



import re
import statistics
import logging

import datetime
from datetime import timezone
from fontTools import unicodedata
from abc import abstractmethod

from openpecha.core.annotation import Page, Span
from openpecha.core.annotations import Language, OCRConfidence
from openpecha.core.layer import Layer, LayerEnum, OCRConfidenceLayer
from openpecha.core.metadata import InitialPechaMetadata, InitialCreationType, LicenseType, Copyright_copyrighted, Copyright_unknown, Copyright_public_domain
from openpecha.core.pecha import OpenPechaFS
from openpecha.formatters import BaseFormatter
from openpecha.core import ids
from openpecha import __version__

ANNOTATION_MINIMAL_LEN = 20
ANNOTATION_MINIMAL_CONFIDENCE = 0.8
ANNOTATION_MAX_LOW_CONF_PER_PAGE = 10
NO_SPACE_AFTER_PATTERN = re.compile(r"(?:\s|[༌་])$")
# mapping between script tags detected from the text and language recorded
# in the Language layer
DEFAULT_SCRIPT_TO_LANG_MAPPING = {
    "Tibt": "bo",
    "Deva": "sa-Deva",
    "Hani": "zh",
    "Hans": "zh",
    "Hant": "zh",
    "Latn": "en",
    "Mong": "mn-Mong",
    "Newa": "sa-Newa",
    "Soyo": "mn-Soyo"
}
NO_LANG = ""
UNKNOWN_LANG = "und"
# Unicode character categories taken into account when computing width:
UNICODE_CHARCAT_FOR_WIDTH = ["Ll", "Lu", "Lo", "Nd", "No", "Nl", "Lt"]
UNICODE_CHARCAT_NOT_NOISE = ["Ll", "Lu", "Lo", "Nd", "No", "Nl", "Lt"]
SAME_LINE_RATIO_THRESHOLD = 0.2

class BBox:
    def __init__(self, x1: int, x2: int, y1: int, y2: int, angle: int = None, text: str = None, confidence: float = None, language: str = NO_LANG):
        self.text = text
        self.x1 = x1
        self.x2 = x2
        self.y1 = y1
        self.y2 = y2
        self.angle = angle
        self.confidence = confidence
        self.language = language
        self.mid_y = (y1 + y2) / 2
        self.mid_x = (x1 + x2) / 2
    
    def get_height(self):
        return self.y2 - self.y1

    def get_width(self):
        return self.x2 - self.x1
    
    def get_angle(self):
        """
        Returns the angle of the BBox. The value is either None (when no angle can be determined)
        or the value of the clockwise rotation, in positive degrees (the value must be between 0 and 359).
        A value of 0 represents straight characters.
        """
        return self.angle

    def get_box_orientation(self):
        width = self.x2 - self.x1
        length = self.y2 - self.y1
        if width > length:
            return "landscape"
        else:
            return "portrait"

    def get_y_mid(self):
        return self.mid_y

    def get_centriod(self):
        return [self.mid_x, self.mid_y]

    def to_debug_str(self):
        return "'%s' [%d,%d,%d,%d]" % (self.text, self.x1, self.x2, self.y1, self.y2)


class OCRFileProvider():
    def __init__(self, ocr_import_info: str):
        self.ocr_import_info = ocr_import_info

    def get_image_list(self, image_group_id: str):
        # to be implemented by sub classes
        # must return a simple list of image ids
        pass

    def get_source_info(self):
        # to be implemented by sub classes
        # must return an dict in the same format as buda.get_buda_scan_info
        pass

    def get_image_data(self, image_group_id, image_id):
        # to be implemented by sub classes
        # must return a list of BBox
        pass

class OCRFormatter(BaseFormatter):
    """
    General OpenPecha Formatter for OCR, must 
    """

    def __init__(self, output_path=None, metadata=None):
        super().__init__(output_path, metadata)
        self.n_page_breaker_char = 3
        self.page_break = "\n" * self.n_page_breaker_char
        self.base_text = []
        self.low_conf_ann_text = ""
        self.base_meta = {}
        self.cur_base_word_confidences = []
        self.word_confidences = []
        self.bdrc_scan_id = None
        self.metadata = {}
        self.default_language = None
        self.source_info = {}
        self.remove_non_character_lines = True
        self.create_language_layer = True
        self.ocr_confidence_threshold = ANNOTATION_MINIMAL_CONFIDENCE
        self.language_annotation_min_len = ANNOTATION_MINIMAL_LEN
        self.remove_rotated_boxes = True
        self.remove_duplicate_symbols = True
        self.script_to_lang_map = DEFAULT_SCRIPT_TO_LANG_MAPPING
        self.max_low_conf_per_page = ANNOTATION_MAX_LOW_CONF_PER_PAGE

    def text_preprocess(self, text):
        return text


    def get_avg_bbox_height(self, bboxes):
        """Calculate the average height of bounding bboxs in page

        Args:
            bboxes (list): list of boundingbboxs

        Returns:
            float: average height of bounding ploys
        """
        height_sum = 0
        bboxeswidth = 0
        for bbox in bboxes:
            # weigh by number of characters
            bboxwidth = bbox.get_width()
            bboxeswidth += bboxwidth
            height_sum += bbox.get_height()*bboxwidth
        avg_height = height_sum / bboxeswidth
        logging.debug("average bbox height: %f (%d, %d)", avg_height, height_sum, bboxeswidth)
        return avg_height

    def is_on_same_line(self, prev_bbox, bbox, y_diff_threshold):
        """Check if bounding bbox is in same line as previous bounding bbox
        a threshold to check the conditions set to 10 but it can varies for pecha to pecha

        Args:
            prev_bbox (dict): previous bounding bbox
            bbox (dict): current bounding bbox
            avg_height (float): average height of all the bounding bboxs in page

        Returns:
            boolean: true if bouding bbox is in same line as previous bounding bbox else false
        """
        return (bbox.mid_y - prev_bbox.mid_y) < y_diff_threshold

    def get_bbox_lines(self, bboxes):
        """Return list of lines in page using bounding bboxs of page

        Args:
            bboxes (list): list of all the bounding bboxs

        Returns:
            list: list of lines in page
        """
        lines = []
        cur_line_bboxs = []
        prev_bbox = bboxes[0]
        avg_line_height = self.get_avg_bbox_height(bboxes)
        y_diff_threshold = int(avg_line_height * self.same_line_ratio_threshold)
        for bbox in bboxes:
            if self.is_on_same_line(prev_bbox, bbox, y_diff_threshold):
                cur_line_bboxs.append(bbox)
            else:
                lines.append(cur_line_bboxs)
                cur_line_bboxs  = []
                cur_line_bboxs.append(bbox)
            prev_bbox = bbox
        if cur_line_bboxs:
            lines.append(cur_line_bboxs)
        return lines


    def get_bbox_sorted_on_y(self, bbox_centriods):
        """Sort bounding bboxs centriod base on y coordinates

        Args:
            bbox_centriods (list): list of centriod coordinates

        Returns:
            list: list centriod coordinated sorted base on y coordinate
        """
        bboxes_sorted_on_y = sorted(bbox_centriods , key=lambda k: [k[1]])
        return bboxes_sorted_on_y
    
    def sort_line_and_remove_duplicates(self, line, bboxes):
        """it sorted the curr line using x centriod and
        use that sorted list of centriod, use the centrod co-ordinates
        to get the bboxes object to get the area of the box and 
        remove it if the area overlap or intersection over union percentage.

        Args:
            line (list): list of centriod co-ordinates
            bboxes (dict): dict of bbox object with centriod co-ordinates as key

        Returns:
            list: sorted new line with removed overlaps or duplicates
        """
        sorted_line = sorted(line, key=lambda k: [k[0]])
        if not self.remove_duplicate_symbols:
            return sorted_line
        new_line = []
        prev_x2 = -1
        orig_str = ""
        for i, bbox in enumerate(sorted_line):
            curr_bbox = bboxes[f"{bbox[0]},{bbox[1]}"]
            if curr_bbox.x1 <= prev_x2:
                # check if there is overlap with a previous box on this line:
                duplicate = False
                for j in reversed(range(0, i)):
                    bbox_j = bboxes[f"{sorted_line[j][0]},{sorted_line[j][1]}"]
                    if curr_bbox.x1 > bbox_j.x2:
                        continue
                    if curr_bbox.x2 < bbox_j.x1:
                        break
                    overlap = min(bbox_j.x2, curr_bbox.x2) - max(bbox_j.x1, curr_bbox.x1)
                    if overlap > 0.6 * (curr_bbox.x2 - curr_bbox.x1):
                        duplicate = True
                        logging.debug("remove duplicate symbol ", bbox_j.text)
                        break
                if duplicate:
                    continue
                else:
                    new_line.append(bbox)
            else:
                new_line.append(bbox)
            prev_x2 = curr_bbox.x2
        return new_line
                    

    def get_bbox_sorted_on_x(self, bboxes_sorted_on_y, avg_box_height, bboxes):
        """Groups box belonging in same line using average height and sort the grouped boxes base on x coordinates

        Args:
            bboxes_sorted_on_y (list): list of centriod
            avg_box_height (float): average boxes height

        Returns:
            list: list of sorted centriod
        """
        prev_bbox = bboxes_sorted_on_y[0]
        lines = []
        cur_line = []
        sorted_bboxes = []
        for bbox in bboxes_sorted_on_y:
            if abs(bbox[1]-prev_bbox[1]) < avg_box_height/10:
                cur_line.append(bbox)
            else:
                lines.append(cur_line)
                cur_line = []
                cur_line.append(bbox)
            prev_bbox = bbox
        if cur_line:
            lines.append(cur_line)
        for line in lines:
            sorted_line = self.sort_line_and_remove_duplicates(line, bboxes)
            # sorted_line = sorted(line, key=lambda k: [k[0]])
            for bbox in sorted_line:
                sorted_bboxes.append(bbox)
        return sorted_bboxes

    def sort_bboxes(self, main_region_bboxes):
        """Sort the bounding bboxs

        Args:
            main_region_bboxes (list): list of bounding bboxs

        Returns:
            list: sorted bounding bbox list
        """
        bboxes = {}
        bbox_centriods = []
        avg_box_height = self.get_avg_bbox_height(main_region_bboxes)
        for bbox in main_region_bboxes:
            centroid = bbox.get_centriod()
            bboxes[f"{centroid[0]},{centroid[1]}"] = bbox
            bbox_centriods.append(centroid)
        sorted_bboxes = []
        sort_on_y_bboxs = self.get_bbox_sorted_on_y(bbox_centriods)
        sorted_bbox_centriods = self.get_bbox_sorted_on_x(sort_on_y_bboxs, avg_box_height, bboxes)
        for bbox_centriod in sorted_bbox_centriods:
            sorted_bboxes.append(bboxes[f"{bbox_centriod[0]},{bbox_centriod[1]}"])
        return sorted_bboxes

    def has_space_attached(self, symbol):
        # must be overriden if the format has that kind of information
        return False

    def populate_confidence(self, bboxes):
        """Populate confidence of bounding bboxs of pecha level and image group level

        Args:
            bboxes (list): list of bounding bboxs
        """
        for bbox in bboxes:
            self.word_confidences.append(float(bbox.confidence))
            self.cur_base_word_confidences.append(float(bbox.confidence))

    def bbox_can_have_space_after(self, bbox):
        if NO_SPACE_AFTER_PATTERN.search(bbox.text):
            return False
        return True
        
    def has_space_after(self, cur_bbox, next_bbox, avg_char_width):
        """Checks if there is space between two bbox or not if yes returns a space bbox bbox object

        Args:
            cur_bbox (bbox): first bbox
            next_bbox (bbox): second bbox
            avg_char_width (float): avg width of char in page

        Returns:
            bbox or none: space bbox object
        """
        if next_bbox.x1 - cur_bbox.x2 > avg_char_width*0.75 and self.bbox_can_have_space_after(cur_bbox):
            space_box = BBox(
                cur_bbox.x2,
                next_bbox.x1,
                cur_bbox.y1, # the y coordinates are kind of arbitrary
                cur_bbox.y2,
                angle=0,
                text=" ",
                confidence=None,
                language=None
            )
            return space_box
        return None
    
    def insert_space_bbox(self, bboxes, avg_char_width):
        """Inserts space bounding bbox if missing

        Args:
            bboxes (list): list of bbox objects
            avg_char_width (float): avg width of char

        Returns:
            list: list of bbox objects
        """
        new_bboxes = []
        for bbox_walker, cur_bbox in enumerate(bboxes):
            if bbox_walker == len(bboxes)-1:
                new_bboxes.append(cur_bbox)
            else:
                next_bbox = bboxes[bbox_walker+1]
                space_bbox = self.has_space_after(cur_bbox, next_bbox, avg_char_width)
                if space_bbox:
                    new_bboxes.append(cur_bbox)
                    new_bboxes.append(space_bbox)
                else:
                    new_bboxes.append(cur_bbox)
        return new_bboxes

    def get_main_script_tag(text: str):
        """
        return the ISO 15924 tag of the main script used
        in a string
        """
        halflen = int(len(text) / 2)
        scripts = {}
        maxcharscript_nbchars = 0
        maxcharscript = "Zyyy"
        for i, c in enumerate(text):
            cscript = unicodedata.script(c)
            if cscript not in scripts:
                scripts[cscript] = 0
            script_nbchars = scripts[cscript] + 1
            scripts[cscript] = script_nbchars
            if script_nbchars > maxcharscript_nbchars:
                maxcharscript_nbchars = script_nbchars
                maxcharscript = cscript
            if maxcharscript_nbchars > halflen:
                return maxcharscript
        return maxcharscript


    def get_main_language_code(self, text):
        """returns the language tag for the text

        Args:
            string (str): text from wordbox.text
        """
        main_script = OCRFormatter.get_main_script_tag(text)
        if main_script in self.script_to_lang_map:
            return self.script_to_lang_map[main_script]
        if main_script == "Zyyy" or main_script == "Zxxx":
            return NO_LANG
        return UNKNOWN_LANG

    def is_noise(self, text):
        for c in text:
            if unicodedata.category(c) in UNICODE_CHARCAT_NOT_NOISE:
                return False
        return True

    def add_language(self, bbox, bbox_start_cc, state):
        bbox_lang = bbox.language
        previous_ann = state["latest_language_annotation"]
        bbox_end_cc = state["base_layer_len"] # by construction
        if previous_ann is not None:
            # if bbox has the same language as the latest annotation, we just lengthen the previous
            # annotation to include this bbox:
            if bbox_lang == previous_ann['lang'] or not bbox_lang:
                previous_ann["end"] = bbox_end_cc
                return
            # if bbox is the default language, we just conclude the previous annotation
            if bbox_lang == self.default_language:
                state["latest_language_annotation"] = None
                return
            # else, we create a new annotation
            annotation = {"start": bbox_start_cc, "end": bbox_end_cc, "lang": bbox_lang}
            state["language_annotations"].append(annotation)
            state["latest_language_annotation"] = annotation
            return
        # if there's no previous annotation and language is the default language, return
        if bbox_lang == self.default_language or bbox_lang == NO_LANG or bbox_lang == UNKNOWN_LANG or not bbox_lang:
            return
        # if there's no previous annotation and language is not the default, we create an annotation
        annotation = {"start": bbox_start_cc, "end": bbox_end_cc, "lang": bbox_lang}
        state["language_annotations"].append(annotation)
        state["latest_language_annotation"] = annotation

    def add_low_confidence(self, bbox, bbox_start_cc, state):
        if bbox.confidence is None:
            return
        if bbox.confidence > self.ocr_confidence_threshold:
            state["latest_low_confidence_annotation"] = None
            return
        bbox_end_cc = state["base_layer_len"] # by construction
        if state["latest_low_confidence_annotation"] is not None:
            # average of the confidence indexes, weighted by character length
            state["latest_low_confidence_annotation"]["weights"].append((bbox_end_cc - bbox_start_cc, bbox.confidence))
            state["latest_low_confidence_annotation"]["end"] = bbox_end_cc
        else:
            annotation = {"start": bbox_start_cc, "end": bbox_end_cc, 
                "weights": [(bbox_end_cc - bbox_start_cc, bbox.confidence)]}
            state["page_low_confidence_annotations"].append(annotation)
            state["latest_low_confidence_annotation"] = annotation

    def bbox_line_has_characters(self, bbox_line):
        for bbox in bbox_line:
            if bbox.language != UNKNOWN_LANG and not self.is_noise(bbox.text):
                return True
        if logging.DEBUG >= logging.root.level:
            line = ''
            for bbox in bbox_line:
                line += bbox.text
            logging.debug("ignoring line '%s', detected as noise", line)
        return False

    def build_page(self, bboxes, image_number, image_filename, state, avg_char_width=None):
        sorted_bboxes = self.sort_bboxes(bboxes)
        bbox_lines = self.get_bbox_lines(sorted_bboxes)
        page_start_cc = state["base_layer_len"]
        page_word_confidences = []
        for bbox_line in bbox_lines:
            if self.remove_non_character_lines and not self.bbox_line_has_characters(bbox_line):
                continue
            if avg_char_width:
                bbox_line = self.insert_space_bbox(bbox_line, avg_char_width)
            for bbox in bbox_line:
                state["base_layer"] += bbox.text
                start_cc = state["base_layer_len"]
                state["base_layer_len"] += len(bbox.text)
                if bbox.confidence is not None:
                    state["word_confidences"].append(float(bbox.confidence))
                    page_word_confidences.append(float(bbox.confidence))
                    self.add_low_confidence(bbox, start_cc, state)
                self.add_language(bbox, start_cc, state)
            # adding a line break at the end of a line
            state["base_layer"] += "\n"
            state["base_layer_len"] += 1
        # if the whole page is below the min confidence level, we just add one
        # annotation for the page instead of annotating each word
        if page_word_confidences:
            mean_page_confidence = statistics.mean(page_word_confidences)
        else:
            mean_page_confidence = 0
        nb_below_threshold = len(state["page_low_confidence_annotations"])
        if mean_page_confidence < self.ocr_confidence_threshold or nb_below_threshold > self.max_low_conf_per_page:
            state["low_confidence_annotations"][self.get_unique_id()] = OCRConfidence(
                span=Span(start=page_start_cc, end=state["base_layer_len"]), 
                confidence=mean_page_confidence, nb_below_threshold=nb_below_threshold if nb_below_threshold else None)
        else:
            self.merge_page_low_confidence_annotations(state["page_low_confidence_annotations"], state["low_confidence_annotations"])
            state["page_low_confidence_annotations"] = []
        # add pagination annotation:
        state["pagination_annotations"][self.get_unique_id()] = Page(
            span=Span(start=page_start_cc, end=state["base_layer_len"]), 
            imgnum=image_number, 
            reference=image_filename)
        # adding another line break at the end of a page
        state["base_layer"] += "\n"
        state["base_layer_len"] += 1

    def confidence_index_from_weighted_list(self, weights):
        sum_weights = 0
        confidence_sum = 0
        for weight, confidence in weights:
            sum_weights += weight
            confidence_sum += weight * confidence
        return confidence_sum / sum_weights

    def merge_page_low_confidence_annotations(self, annotation_list_src, annotations_dst):
        for annotation in annotation_list_src:
            avg_confidence = self.confidence_index_from_weighted_list(annotation["weights"])
            annotation_obj = OCRConfidence(
                span = Span(start=annotation["start"], end=annotation["end"]),
                confidence = avg_confidence)
            annotations_dst[self.get_unique_id()] = annotation_obj

    def merge_short_language_annotations(self, annotation_list):
        annotations = {}
        previous_annotation = None
        # annotation list is in span order
        for annotation in annotation_list:
            if annotation['end'] - annotation['start'] < self.language_annotation_min_len:
                if previous_annotation is not None and annotation['start'] - previous_annotation.span.end < self.language_annotation_min_len:
                    previous_annotation.span.end = annotation['end']
                continue
            if previous_annotation is not None and annotation["lang"] == previous_annotation.language:
                if annotation['start'] - previous_annotation.span.end < self.language_annotation_min_len:
                    previous_annotation.span.end = annotation['end']
                    continue
            previous_annotation = Language(
                span = Span(start=annotation["start"], end=annotation["end"]),
                language = annotation["lang"])
            annotations[self.get_unique_id()] = previous_annotation
        return annotations
    
    @abstractmethod
    def get_bboxes_for_page(self, image_group_id, image_filename):
        # needs to be implemented by inheriting classes
        pass

    def build_base(self, image_group_id):
        """ The main function that takes the OCR results for an entire volume
            and creates its base and layers
        """
        image_list = self.data_provider.get_image_list(image_group_id)
        state = {
            "base_layer_len": 0,
            "base_layer": "",
            "low_confidence_annotations": {},
            "language_annotations": [],
            "pagination_annotations": {},
            "word_confidences": [],
            "latest_language_annotation": None,
            "latest_low_confidence_annotation": None,
            "page_low_confidence_annotations": []
        }
        for image_number, image_filename in enumerate(image_list):
            # enumerate starts at 0 but image numbers start at 1
            bboxes, avg_char_width = self.get_bboxes_for_page(image_group_id, image_filename)
            if bboxes:
                self.build_page(bboxes, image_number+1, image_filename, state, avg_char_width)
        layers = {}
        if state["pagination_annotations"]:
            layer = Layer(annotation_type=LayerEnum.pagination, annotations=state["pagination_annotations"])
            layers[LayerEnum.pagination] = layer
        if state["language_annotations"]:
            annotations = self.merge_short_language_annotations(state["language_annotations"])
            layer = Layer(annotation_type=LayerEnum.language, annotations=annotations)
            layers[LayerEnum.language] = layer
        if state["low_confidence_annotations"]:
            layer = OCRConfidenceLayer(
                confidence_threshold=self.ocr_confidence_threshold, 
                annotations=state["low_confidence_annotations"]
            )
            layers[LayerEnum.ocr_confidence] = layer
        return state["base_layer"], layers, state["word_confidences"]

    def get_copyright_and_license_info(self, bdata):
        if "copyright_status" not in bdata["source_metadata"]:
            return {}, None
        cs = bdata["source_metadata"]["copyright_status"]
        if cs == "http://purl.bdrc.io/resource/CopyrightPublicDomain":
            return Copyright_public_domain, LicenseType.CC0
        if cs == "http://purl.bdrc.io/resource/CopyrightUndetermined":
            return Copyright_unknown, None
        return Copyright_copyrighted, LicenseType.UNDER_COPYRIGHT
            
    def get_metadata(self, pecha_id: str = None, ocr_import_info=None) -> InitialPechaMetadata:
        source_metadata = {
            "id": f"http://purl.bdrc.io/resource/{self.bdrc_scan_id}",
            "title": "",
            "author": "",
        }
        copyright = {}
        license = None
        if self.source_info is not None:
            source_metadata = self.source_info["source_metadata"]
            copyright, license = self.get_copyright_and_license_info(self.source_info)

        parser_link = ocr_import_info["parser_link"] if "parser_link" in ocr_import_info else None

        metadata = InitialPechaMetadata(
            id=pecha_id,
            source='https://library.bdrc.io',
            initial_creation_type=InitialCreationType.ocr,
            imported=datetime.datetime.now(timezone.utc),
            last_modified=datetime.datetime.now(timezone.utc),
            parser=parser_link,
            copyright=copyright,
            license=license,
            source_metadata=source_metadata,
            default_language=self.default_language,
            ocr_import_info=ocr_import_info
        )
        return metadata

    def set_base_meta(self, image_group_id, base_file_name, word_confidence_list):
        self.cur_word_confidences = []
        self.base_meta[base_file_name] = {
            "source_metadata": self.source_info["image_groups"][image_group_id],
            "order": self.source_info["image_groups"][image_group_id]["volume_number"],
            "base_file": f"{base_file_name}.txt"
            }
        if word_confidence_list:
            self.base_meta[base_file_name]["statistics"] = {
              "ocr_word_median_confidence_index": statistics.median(word_confidence_list),
              "ocr_word_mean_confidence_index": statistics.mean(word_confidence_list)
            }
    
    def create_opf(self, data_provider, pecha_id = None, opf_options = {}, ocr_import_info = {}):
        """Create opf

        Args:
            data_provider (DataProvider): an instance that will be used to get the necessary data
            pecha_id (str): pecha id
            opf_options (Dict): an object with the following keys:
                create_language_layer: boolean
                language_annotation_min_len: int
                ocr_confidence_threshold: float (use -1.0 for no OCR confidence layer)
                remove_non_character_lines: boolean
                max_low_conf_per_page: int
            ocr_import_info (Dict): an object with the following keys:
                bdrc_scan_id: str
                source: str
                ocr_info: Dict
                batch_id: str
                software_id: str
                expected_default_language: str

        Returns:
            path: opf path
        """

        self.data_provider = data_provider

        self.remove_non_character_lines = opf_options["remove_non_character_lines"] if "remove_non_character_lines" in opf_options else True
        self.remove_rotated_boxes = opf_options["remove_rotated_boxes"] if "remove_rotated_boxes" in opf_options else True
        self.create_language_layer = opf_options["create_language_layer"] if "create_language_layer" in opf_options else True
        self.ocr_confidence_threshold = opf_options["ocr_confidence_threshold"] if "ocr_confidence_threshold" in opf_options else ANNOTATION_MINIMAL_CONFIDENCE
        self.language_annotation_min_len = opf_options["language_annotation_min_len"] if "language_annotation_min_len" in opf_options else ANNOTATION_MINIMAL_LEN
        self.max_low_conf_per_page = opf_options["max_low_conf_per_page"] if "max_low_conf_per_page" in opf_options else ANNOTATION_MAX_LOW_CONF_PER_PAGE
        self.script_to_lang_map = opf_options["script_to_lang_map"] if "script_to_lang_map" in opf_options else DEFAULT_SCRIPT_TO_LANG_MAPPING
        self.same_line_ratio_threshold = opf_options["same_line_ratio_threshold"] if "same_line_ratio_threshold" in opf_options else SAME_LINE_RATIO_THRESHOLD
        self.remove_duplicate_symbols = opf_options["remove_duplicate_symbols"] if "remove_duplicate_symbols" in opf_options else True

        ocr_import_info["op_import_options"] = opf_options
        ocr_import_info["op_import_version"] = __version__
        
        # self._build_dirs(None, id_=pecha_id)

        # if the bdrc scan id is not specified, we assume it's the directory namepecha_id
        self.bdrc_scan_id = self.data_provider.bdrc_scan_id
        self.source_info = self.data_provider.get_source_info()
        self.default_language = "bo" if "expected_default_language" not in ocr_import_info else ocr_import_info["expected_default_language"]
        self.default_language = "bo"
        if "expected_default_language" in ocr_import_info:
            self.default_language = ocr_import_info["expected_default_language"]
        elif "languages" in self.source_info and self.source_info["languages"]:
            self.default_language = self.source_info["languages"][0]
        pecha_id = ids.get_initial_pecha_id() if pecha_id is None else pecha_id
        self.metadata = self.get_metadata(pecha_id, ocr_import_info)
        pecha = OpenPechaFS(metadata=self.metadata,
                            path=self.output_path / pecha_id / f"{pecha_id}.opf",
                            pecha_id=pecha_id)
        total_word_confidence_list = []
    
        for image_group_id, _ in self.source_info["image_groups"].items():
            base_id = image_group_id
            base_text, layers, word_confidence_list = self.build_base(image_group_id)
            pecha.bases[base_id] = base_text
            pecha.layers[base_id] = layers
            self.set_base_meta(image_group_id, base_id, word_confidence_list)
            total_word_confidence_list += word_confidence_list

        # we add the rest to metadata:
        pecha.meta.bases = self.base_meta
        if total_word_confidence_list:
            pecha.meta.statistics = {
                # there are probably more efficient ways to compute those
                "ocr_word_mean_confidence_index": statistics.mean(total_word_confidence_list),
                "ocr_word_median_confidence_index": statistics.median(total_word_confidence_list)
            }
        pecha.save()

        return pecha



"""Utilities functions
"""
import gzip
import io
import shutil
import urllib
from collections import defaultdict
from pathlib import Path
from typing import Dict

import requests
import yaml
from git import Repo
from git.cmd import GitCommandError

from openpecha import config
from openpecha.core import metadata
from openpecha.core.layer import LayerEnum
from openpecha.exceptions import PechaNotFound
from openpecha.github_utils import create_release
from openpecha.storages import GithubStorage, setup_auth_for_old_repo

INFO = "[INFO] {}"

# use yaml.CSafeLoader / if available but don't crash if it isn't
try:
    yaml_loader = yaml.CSafeLoader
except (ImportError, AttributeError):
    yaml_loader = yaml.SafeLoader

try:
    yaml_dumper = yaml.CSafeDumper
except (ImportError, AttributeError):
    yaml_dumper = yaml.SafeDumper


def simple_enum_to_yaml(representer, node):
    return representer.represent_data(node.value)


yaml_dumper.add_multi_representer(LayerEnum, simple_enum_to_yaml)
yaml_dumper.add_multi_representer(metadata.InitialCreationType, simple_enum_to_yaml)
yaml_dumper.add_multi_representer(metadata.CopyrightStatus, simple_enum_to_yaml)
yaml_dumper.add_multi_representer(metadata.LicenseType, simple_enum_to_yaml)


def gzip_str(string_):
    # taken from https://gist.github.com/Garrett-R/dc6f08fc1eab63f94d2cbb89cb61c33d
    out = io.BytesIO()

    with gzip.GzipFile(fileobj=out, mode="w") as fo:
        fo.write(string_.encode())

    bytes_obj = out.getvalue()
    return bytes_obj


def _mkdir(path):
    if path.is_dir():
        return path
    path.mkdir(exist_ok=True, parents=True)
    return path


def ocr_result_input(path):
    return path


def create_release_with_assets(path):
    asset_paths = []
    for asset_path in (path / "releases").iterdir():
        shutil.make_archive(asset_path, "zip", asset_path)
        asset_paths.append(f"{str(asset_path)}.zip")

    create_release(path.name, asset_paths=asset_paths)


class Vol2FnManager:
    def __init__(self, metadata):
        self.name = "vol2fn"
        self.vol_num = 0
        self.vol2fn = self._get_vol2fn(metadata)
        self.fn2vol = {fn: vol for vol, fn in self.vol2fn.items()}

    def _get_vol2fn(self, metadata):
        if self.name in metadata:
            return metadata[self.name]
        else:
            return defaultdict(dict)

    def get_fn(self, vol):
        return self.vol2fn.get(vol)

    def get_vol_id(self, fn):
        vol_id = self.fn2vol.get(fn)
        if vol_id:
            return vol_id
        else:
            self.vol_num += 1
            vol_id = f"v{self.vol_num:03}"
            self.vol2fn[vol_id] = fn
            return vol_id


def dump_yaml(data: Dict, output_fn: Path) -> Path:
    with output_fn.open("w", encoding="utf-8") as fn:
        yaml.dump(
            data,
            fn,
            default_flow_style=False,
            sort_keys=False,
            allow_unicode=True,
            Dumper=yaml_dumper,
        )
    return output_fn


def load_yaml(fn: Path) -> None:
    return yaml.load(fn.open(encoding="utf-8"), Loader=yaml_loader)


def load_yaml_str(s: str) -> None:
    return yaml.load(s, Loader=yaml_loader)


def _eval_branch(repo, branch):
    """return default branch as fallback branch."""
    if branch in repo.refs or f"origin/{branch}" in repo.refs:
        return branch
    elif "main" in repo.refs:
        return "main"
    else:
        return "master"


def download_pecha(pecha_id, out_path=None, needs_update=True, branch="main") -> Path:
    """Download pecha from github

    Note: If pecha is already downloaded before and needs_update,
          it will update the pecha by pulling the remote `branch`.

    Args:
        pecha_id (str): pecha id
        out_path (str, optional): path to download pecha. Defaults to None.
        needs_update (bool, optional): if True, update pecha. Defaults to True.
        branch (str, optional): branch name. Defaults to "main".

    Returns:
        Path: pecha path
    """
    # resolve defaults kwargs
    needs_update = needs_update if needs_update is not None else True
    branch = branch if branch is not None else "main"
    storage = GithubStorage()

    # create pecha path
    if out_path:
        out_path = Path(out_path)
        out_path.mkdir(exist_ok=True, parents=True)
        pecha_path = out_path / pecha_id
    else:
        pecha_path = config.PECHAS_PATH / pecha_id

    if pecha_path.is_dir():
        repo = Repo(str(pecha_path))
        branch = _eval_branch(repo, branch)
        repo.git.checkout(branch)
        if needs_update:
            repo.git.config("pull.rebase", "false")
            repo.git.pull("origin", branch)
    else:
        pecha_url = storage.get_authenticated_repo_remote_url(pecha_id)
        try:
            Repo.clone_from(pecha_url, str(pecha_path))
        except GitCommandError:
            raise PechaNotFound(f"Pecha with id {pecha_id} doesn't exist")
        repo = Repo(str(pecha_path))
        branch = _eval_branch(repo, branch)
        repo.git.checkout(branch)

    # setup auth
    setup_auth_for_old_repo(repo, org=storage.org_name, token=storage.token)

    return pecha_path


def download_pecha_assets(pecha_id: str, asset_type: str, download_dir: Path):
    """Download pecha assets from latest release

    Args:
        pecha_id (str): pecha id
        asset_type (str): asset type can be src web page or ocr output
        download_dir (Path): directory where you want to download the assets

    Returns:
        Path: zip file path of downloaded asset
    """

    response = requests.get(
        f"https://api.github.com/repos/OpenPecha-data/{pecha_id}/releases/latest"
    )
    res = response.json()
    for asset in res["assets"]:
        if asset["name"] == f"{asset_type}.zip":
            asset_download_url = asset["browser_download_url"]
            break
    f = urllib.request.urlopen(asset_download_url)
    assets = f.read()
    zip_asset_file_path = download_dir / f"{pecha_id}_{asset_type}.zip"

    zip_asset_file_path.write_bytes(assets)

    return zip_asset_file_path



import re

from pathlib import Path
from openpecha.core.annotations import TranscriptionTimeSpan, Span
from openpecha.formatters.formatter import BaseFormatter
from openpecha.core.layer import TranscriptionTimeSpanLayer, LayerEnum
from openpecha.core.annotations import TranscriptionTimeSpan, Span
from openpecha.core.pecha import OpenPechaFS
from openpecha.core.ids import get_initial_pecha_id, get_base_id
from openpecha.core.metadata import InitialPechaMetadata
import json
from bs4 import BeautifulSoup


def get_json_from_str(text):
    otr_json = json.loads(text)
    return otr_json


def parse_html(html):
    soup = BeautifulSoup(html, "html.parser")
    return soup


def get_all_timestamps(soup):
    return soup.findAll("span", {"class": "timestamp"})


def get_text_between(start, end):
    text = ""
    tag = start.next_element.next_element
    while tag != end:
        if tag.name is None:
            text += tag.text
        tag = tag.next_element
    return text


class OTranscribeFormatter(BaseFormatter):
    """
    oTranscribe formatter for .otr file.
    """

    def __init__(
        self,
        output_path="./output",
        metadata=None,
        media_url=None,
        time_unit="microsecond",
    ):
        super().__init__(output_path, metadata)
        self.media_url = media_url
        self.time_unit = time_unit

    def text_preprocess(self, text):
        otr_json = get_json_from_str(text)
        return otr_json["text"].strip()

    def build_layers(self, text):
        transcription_time_span_layer = TranscriptionTimeSpanLayer(
            media_url=self.media_url,
            time_unit=self.time_unit,
        )

        soup = parse_html(text)
        timestamps = get_all_timestamps(soup)
        base_text = ""
        char_walker = 0
        for start, end in zip(timestamps, timestamps[1:]):
            time_start = float(start["data-timestamp"]) * 10**6
            time_end = float(end["data-timestamp"]) * 10**6
            time_span = Span(start=time_start, end=time_end)
            text_between = get_text_between(start, end).strip()
            text_end = char_walker + len(text_between)
            text_span = Span(start=char_walker, end=text_end)
            char_walker += len(text_between) + 1
            transcription_time_span_layer.set_annotation(
                TranscriptionTimeSpan(span=text_span, time_span=time_span)
            )
            base_text += text_between + "\n"
        # \xa0 is non-breaking space. Replace it with space.
        base_text = base_text.replace("\xa0", " ")
        return {
            LayerEnum.transcription_time_span: transcription_time_span_layer
        }, base_text

    def create_opf_from_dir(self, input_path):
        input_path = Path(input_path)
        if not input_path.is_dir():
            raise TypeError("Only directories are allowed")
        if not len(list(input_path.glob("*.otr"))):
            raise Exception("No .otr file found")

        pecha_id = get_initial_pecha_id()
        pecha_path = self.output_path / f"{pecha_id}/{pecha_id}.opf"
        pecha_meta = InitialPechaMetadata(id=pecha_id)
        pecha = OpenPechaFS(path=pecha_path, pecha_id=pecha_id)
        pecha._meta = pecha_meta

        for otr_file in input_path.glob("*.otr"):
            m_text = self.get_input(otr_file)
            layers, base_text = self.build_layers(m_text)

            base_id = get_base_id()

            pecha.set_base(
                content=base_text,
                base_name=base_id,
                metadata={
                    "source_metadata": {
                        "title": "test",
                        "author": "sonam",
                        "id": pecha_id,
                    },
                    "base_file": f"{base_id}.txt",
                },
            )
            pecha.set_layer(
                base_name=base_id, layer=layers[LayerEnum.transcription_time_span]
            )
        return pecha

    def create_opf(self, input_path):
        input_path = Path(input_path)
        if input_path.is_dir():
            return self.create_opf_from_dir(input_path)
        pecha_id = get_initial_pecha_id()
        m_text = self.get_input(input_path)
        layers, base_text = self.build_layers(m_text)
        pecha_path = self.output_path / f"{pecha_id}/{pecha_id}.opf"
        pecha_meta = InitialPechaMetadata(id=pecha_id)
        pecha = OpenPechaFS(path=pecha_path, pecha_id=pecha_id)
        base_id = get_base_id()

        pecha._meta = pecha_meta
        pecha.set_base(
            content=base_text,
            base_name=base_id,
            metadata={
                "source_metadata": {"title": "test", "author": "sonam", "id": pecha_id},
                "base_file": f"{base_id}.txt",
            },
        )
        pecha.set_layer(
            base_name=base_id, layer=layers[LayerEnum.transcription_time_span]
        )
        return pecha



# from .empty import EmptyEbook
from .formatter import BaseFormatter
from .hfml import HFMLFormatter
from .ocr import GoogleVisionFormatter
from .pedurma import PedurmaFormatter
from .tsadra import TsadraFormatter
from .otranscribe import OTranscribeFormatter



import logging
import shutil
import traceback
from pathlib import Path

import click
from tqdm import tqdm

import openpecha
from openpecha import config
from openpecha.alignment.tmx import create_opf as alignment
from openpecha.blupdate import PechaBaseUpdate
from openpecha.core.pecha import OpenPechaFS
from openpecha.corpus.quality import NonWordsCounter
from openpecha.formatters import GoogleVisionFormatter, HFMLFormatter
from openpecha.serializers import EpubSerializer, HFMLSerializer
from openpecha.utils import download_pecha as download_pecha_util

OP_PATH = config.BASE_PATH
config = {
    # Github
    "OP_CATALOG_URL": "https://raw.githubusercontent.com/OpenPoti/openpecha-catalog/master/data/catalog.csv",
    "OP_ORG": "https://github.com/OpenPecha",
    # Local
    "OP_DATA_PATH": (OP_PATH / "data").resolve(),
    "OP_PECHAS_PATH": config.PECHAS_PATH.resolve(),
    "OP_CATALOG_PATH": (OP_PATH / "data" / "catalog.csv").resolve(),
    "CONFIG_PATH": (OP_PATH / "config").resolve(),
    "DATA_CONFIG_PATH": (OP_PATH / "config" / "data_config").resolve(),
}

ERROR = "[ERROR] {}"
INFO = "[INFO] {}"


def get_pecha_id(n):
    return f"P{int(n):06}"


@click.group()
@click.version_option(version=openpecha.__version__)
def cli():
    pass


@cli.command()
@click.argument("pecha_id")
@click.option("--out", "-o", help="Directory to save the pecha")
@click.option("--branch", "-b", help="Which branch to download, default is `main`")
def download_pecha(**kwargs):
    """
    Download a single pecha.
    """
    pecha_id = kwargs["pecha_id"]
    output_path = kwargs["out"]
    branch = kwargs["branch"]

    click.echo(f"✨ Downloading {pecha_id}️...")

    try:
        pecha_path = download_pecha_util(pecha_id, out_path=output_path, branch=branch)
    except Exception as e:
        click.echo(f"❌ Failed to download {pecha_id}...x")
        click.echo(f"❌ {e}")
        return

    click.echo(f"✅ Download completed")
    click.echo(f"📖 Pecha saved at {pecha_path.resolve()}")


@cli.command()
@click.argument("batch_file_path")
@click.option("--out", "-o", help="Directory to save the pecha")
@click.option("--branch", "-b", help="Which branch to download, default is `main`")
def download_pecha_batch(**kwargs):
    """
    Download a batch of pechas.
    """
    batch_file_path = Path(kwargs["batch_file_path"])
    output_path = kwargs["out"]
    branch = kwargs["branch"]

    assert batch_file_path.is_file()

    # resolve outpout path
    output_path = output_path if output_path else (Path.cwd() / batch_file_path.stem)
    output_path = Path(output_path).resolve()
    output_path.mkdir(exist_ok=True, parents=True)

    click.echo(f"✨ Downloading  the batch {batch_file_path.name}...")

    for pecha_id in tqdm(batch_file_path.read_text().splitlines()):
        if not pecha_id:
            continue

        try:
            download_pecha_util(pecha_id, out_path=output_path, branch=branch)
        except Exception as e:
            click.echo(f"❌ Failed to download {pecha_id}...x")
            click.echo("❌ Error: {e}")
            continue

    click.echo(f"✅ Download completed")
    click.echo(f"📚 Batch saved at {output_path}")


# OpenPecha Formatter
formatter_types = ["ocr", "hfml(default)", "tsadra"]


@cli.command()
@click.option(
    "--name", "-n", type=click.Choice(formatter_types), help="Type of formatter"
)
@click.option("--id", "-i", type=int, help="Id of the pecha")
@click.option("--output_path", "-o", help="output path to store opf pechas")
@click.argument("input_path")
def import_text(**kwargs):
    """
    Import text from a file to OpenPecha format.
    """
    if kwargs["name"] == "ocr":
        formatter = GoogleVisionFormatter(kwargs["output_path"])
    elif kwargs["name"] == "tsadra":
        formatter = HFMLFormatter(kwargs["output_path"])
    else:
        formatter = HFMLFormatter(kwargs["output_path"])

    formatter.create_opf(kwargs["input_path"], kwargs["id"])


export_types = ["hfml(default)", "epub"]


@cli.command()
@click.option(
    "--name", "-n", type=click.Choice(export_types), help="Type of export format"
)
@click.option("--opf_path", "-op")
@click.option("--output_path", "-o")
def export_text(**kwargs):
    """
    Export text from OpenPecha format to a specified file format.
    """

    opf_path = kwargs["opf_path"]
    output_path = kwargs["output_path"]
    # if not opf_path:
    #     opf_path = f'{config["OP_PECHAS_PATH"]}/{pecha_id}/{pecha_id}.opf'

    if kwargs["name"] == "epub":
        serializer = EpubSerializer(opf_path)
    else:
        serializer = HFMLSerializer(opf_path)
    serializer.serialize(output_path)


@cli.command()
@click.argument("pecha_number")
@click.argument("pecha_path")
def update_layers(**kwargs):
    """
    Update all the layers when base has been updated.
    """
    pecha_id = get_pecha_id(kwargs["pecha_number"])
    src_pecha_path = download_pecha_util(pecha_id)

    click.echo(INFO.format(f"Updating base of {pecha_id} ..."))
    src_opf_path = src_pecha_path / f"{pecha_id}.opf"
    dst_opf_path = Path(kwargs["pecha_path"]) / f"{pecha_id}.opf"
    pecha = PechaBaseUpdate(src_opf_path, dst_opf_path)
    pecha.update()


def get_alignment_ids(path):
    path = Path(path)
    return [line for line in path.read_text().splitlines() if line]


alignment_import_types = ["tmx"]


@cli.command()
@click.argument("path")
@click.option("--verbose", "-v", help="verbose", is_flag=True)
@click.option("--debug", "-d", help="debug", is_flag=True)
@click.option(
    "--type",
    "-t",
    type=click.Choice(alignment_import_types),
    help="Type of alignment import format. Default is tmx",
)
@click.option(
    "--tm-path", "-tm", help="path to file containing OpenPecha alignment id's"
)
def import_alignment(**kwargs):
    """
    import existing alignment into OpenPecha
    """
    try:
        from openpecha.alignment.integrations.tx import OPATransifexProject
    except Exception:
        raise ImportError(
            "install with `pip install openpecha[tx]` for transifex dependencies"
        )

    if kwargs["debug"]:
        logging.basicConfig(level=logging.DEBUG)
    elif kwargs["verbose"]:
        logging.basicConfig(level=logging.INFO)
    logging.info("Creating OpenPecha alignment")
    alignment_path = alignment.create_alignment_from_tmx(tmx_path=Path(kwargs["path"]))

    added_tms_path = []
    try:
        logging.info(f"Alignment created at {alignment_path}")
        project = OPATransifexProject(org_slug="esukhia", alignment_path=alignment_path)
        project.create()
        project.import_translation()
        logging.info("Imported translation")
        logging.info("Adding TMs")
        added_tms_path = project.add_tm(
            alignment_ids=get_alignment_ids(kwargs["tm_path"])
        )
    except Exception as e:
        print(traceback.print_exc(), e)
        logging.info("Cleaning up")
        # shutil.rmtree(str(alignment_path))
    finally:
        if alignment_path.is_dir and added_tms_path:
            logging.info("Cleaning up")
            if alignment_path.is_dir():
                shutil.rmtree(str(alignment_path))
            for path in added_tms_path:
                shutil.rmtree(str(path))


@cli.command()
@click.argument("path")
@click.option("-t", "--title", help="title of the text")
@click.option(
    "--tm-path", "-tm", help="path to file containing OpenPecha alignment id's"
)
def new_translation(**kwargs):
    """
    prepare translation project in transifex
    """
    try:
        from openpecha.alignment.integrations.tx import OPATransifexProject
    except Exception:
        raise ImportError(
            "install with `pip install openpecha[tx]` for transifex dependencies"
        )

    alignment_path = alignment.create_alignment_from_source_text(
        text_path=Path(kwargs["path"])
    )
    project = OPATransifexProject(org_slug="esukhia", alignment_path=alignment_path)
    project.create()
    project.start_new_translation()
    project.add_tm(alignment_ids=get_alignment_ids(kwargs["tm_path"]))


@cli.command()
@click.argument("pecha_path")
@click.option("--save", "-s", help="save non-words counts", is_flag=True)
def qc(**kwargs):
    """
    check quality of pecha base text
    """
    pecha = OpenPechaFS(path=kwargs["pecha_path"])

    click.echo("Counting Non Words...")
    non_words_count = NonWordsCounter(empty=True)
    for base_name in pecha.components.keys():
        text = pecha.get_base(base_name)
        non_words_count += NonWordsCounter(text)

    if kwargs["save"]:
        pecha.meta.statistics["total_words"] = non_words_count.total_words
        pecha.meta.statistics["total_non_words"] = non_words_count.total_non_words
        pecha.meta.quality["non_words_ratio"] = non_words_count.non_word_ratio
        pecha.save_meta()
